#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoeuf import api, models


class APIModel(models.Model):
    _name = 'xoeuf.tests.test_api.model'

    @api.from_active_ids(leak_context=True)
    def return_self_ids(self):
        return list(self.ids)

    @api.from_active_ids
    def return_ids_and_call_method(self, ids, methodname):
        return list(self.ids), getattr(self.browse(ids), methodname)()

    @api.from_active_ids(leak_context=True)
    def leaked_return_ids_and_call_method(self, ids, methodname):
        return list(self.ids), getattr(self.browse(ids), methodname)()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# __openerp__
# ---------------------------------------------------------------------
# Copyright (c) 2017 Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can redistribute it and/or modify it under the
# terms of the LICENCE attached (see LICENCE file) in the distribution
# package.
#
# Created on 2017-08-01

{
    'name': 'test_timedelta',
    'author': 'Merchise Autrement [~º/~] and Contributors',
    'description': 'Test for the field.Timedelta extension',
    'depends': ['base', ],
    'installable': True,
    'auto_install': False,
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from enum import IntEnum, Enum
from xoeuf import models, fields


class COLORS(IntEnum):
    Blue = 0
    Red = 1
    Green = 2


class CARS(Enum):
    FORD = object()
    CHEV = object()


class Pax(object):
    single = 1
    double = 2
    __members__ = {'single': single, 'double': double}


class Model(models.Model):
    _name = 'test.enum.model'
    color = fields.Enumeration(COLORS)
    car = fields.Enumeration(CARS)
    pax = fields.Enumeration(Pax)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from datetime import timedelta
from hypothesis import strategies, given
from xoeuf.odoo.tests.common import TransactionCase


timedeltas = strategies.timedeltas()
ALMOST_A_SECOND = timedelta(seconds=1, microseconds=-0.0001)


class TestTimedelta(TransactionCase):
    def setUp(self):
        super(TestTimedelta, self).setUp()
        self.Value = self.env['test.timedelta.value']

    @given(timedeltas)
    def test_create(self, value):
        obj = self.Value.create({
            'delta': value
        })
        id = obj.id
        self.Value.invalidate_cache()
        obj = self.Value.browse(id)
        self.assertAlmostEqual(obj.delta, value, delta=ALMOST_A_SECOND)

    @given(timedeltas)
    def test_create2(self, value):
        obj = self.Value.create({
            'delta': value.total_seconds()
        })
        id = obj.id
        self.Value.invalidate_cache()
        obj = self.Value.browse(id)
        self.assertIsInstance(obj.delta, timedelta)
        self.assertAlmostEqual(obj.delta, value, delta=ALMOST_A_SECOND)

    @given(timedeltas)
    def test_set(self, value):
        obj = self.Value.create({})
        obj.delta = value
        id = obj.id
        self.Value.invalidate_cache()
        obj = self.Value.browse(id)
        self.assertAlmostEqual(obj.delta, value, delta=ALMOST_A_SECOND)

    @given(timedeltas)
    def test_set2(self, value):
        obj = self.Value.create({})
        obj.delta = value.total_seconds()
        id = obj.id
        self.Value.invalidate_cache()
        obj = self.Value.browse(id)
        self.assertIsInstance(obj.delta, timedelta)
        self.assertAlmostEqual(obj.delta, value, delta=ALMOST_A_SECOND)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


def get_modelname(model):
    '''Gets the ORM object's name for a model class/object.

    Main usage::

        self.pool[get_modelname(some_model_class)]

    :param model: Either an object (i.e an instance bound to some database) or
                  the any of the it's class definitions.


    '''
    from xoutil.eight import string_types
    from xoeuf.odoo.models import BaseModel
    from xoeuf.models._proxy import ModelProxy
    if isinstance(model, ModelProxy):
        # Minor hack to support models imported using 'xoeuf.models' stuff
        return model._ModelProxy__model
    if not isinstance(model, BaseModel) and not issubclass(model, BaseModel):
        msg = "Invalid argument '%s' for param 'model'" % model
        raise TypeError(msg)
    result = model._name
    if not result:
        # This is the case of a model class having no _name defined, but then
        # it must have the _inherit and _name is regarded the same by OpenERP.
        result = model._inherit
    assert isinstance(result, string_types), ('Got an invalid name for %r' %
                                              model)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


import json

from xoeuf.odoo import http
from xoeuf.models.proxy import ResUsers as User


class Controller(http.Controller):
    @http.route('/test_proxy_none', type='http', auth='none')
    def count_users_none(self):
        return json.dumps(User.search([], count=True))

    @http.route('/test_proxy_pub', type='http', auth='public')
    def count_users_public(self):
        return json.dumps(User.search([], count=True))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Helpers to generate ``BaseModel.write`` commands.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


class _BaseWriter(object):
    '''Core of the function for `ORMWriter`:class: and `ORMCreator`:class:.

    .. warning:: Currently we don't support nested context managers.

    '''
    def __init__(self, model):
        from xoutil.future.collections import StackedDict
        self.model = model
        self._commands = StackedDict()
        self.result = None

    def _get_field(self, attrname):
        return self.model._fields[attrname]

    __getitem__ = _get_field

    def _is_many2many(self, attrname):
        from xoeuf.odoo.fields import Many2many
        rel_types = (Many2many, )
        try:
            from openerp.osv.fields import many2many
            rel_types += (many2many, )
        except ImportError:
            pass
        return isinstance(self[attrname], rel_types)

    def _is_one2many(self, attrname):
        from xoeuf.odoo.fields import One2many
        # XXX: Don't understand next
        try:
            from openerp.osv.fields import one2many
        except ImportError:
            class one2many(object):
                # I have no instances
                pass
        return isinstance(self[attrname], (one2many, One2many))

    @property
    def commands(self):
        '''The current stack of commands at the current level.

        Since commands could be nested, this will only return the
        top-of-the-stack commands.

        :rtype: dict

        '''
        # TODO: Deal with nested writers properly.
        return self._commands.peek()

    def __enter__(self):
        self._commands.push_level()
        return self

    def __exit__(self, ex_type, exc_value, exc_tb):
        raise NotImplemented

    def create(self, attrname, **values):
        '''Issues a command to create a newly related object with given `values` for
        the many2many/one2many column `attrname`.

        '''
        assert self._is_many2many(attrname) or self._is_one2many(attrname)
        from .orm import CREATE_RELATED
        cmds = self._commands.setdefault(attrname, [])
        cmds.append(CREATE_RELATED(**values))

    def update(self, *replacement, **values):
        '''Issues update commands.

        There are two possible signatures for this method:

        - Keywords only `attrname=value`.  This case is used mostly for
          "plain" columns.  If any `attrname` is a relation it must be a
          many2many, and value must be either None or a list of all related
          objects' ids.

        - Two positional arguments and several keyword arguments, like
          ``update(attrname, id, **values)``.  This is for updating the
          attributes of a single related object.  In this case `attrname` must
          be either a many2many or one2many column.

        '''
        from xoutil.future.types import is_collection
        from .orm import UPDATE_RELATED
        commands = self._commands
        if replacement:
            attrname, id = replacement
            assert self._is_many2many(attrname) or self._is_one2many(attrname)
            cmds = commands.setdefault(attrname, [])
            cmds.append(UPDATE_RELATED(id, **values))
        else:
            for attrname, val in values.items():
                if self._is_many2many(attrname):
                    if val is None:
                        self.forgetall(attrname)
                    elif is_collection(val):
                        self.replace(attrname, *tuple(val))
                    else:
                        raise TypeError("Malformed value '%s' for "
                                        "attribute %s" % (val, attrname))
                elif self._is_one2many(attrname):
                    raise TypeError("'update' does not allow one2many keyword"
                                    "arguments.")
                else:
                    commands[attrname] = val

    def remove(self, attrname, *ids):
        '''Issues several "remove related object" commands for the attribute
        named `attrname`.

        '''
        assert self._is_many2many(attrname) or self._is_one2many(attrname)
        from .orm import REMOVE_RELATED
        cmds = self._commands.setdefault(attrname, [])
        cmds.extend(REMOVE_RELATED(id) for id in ids)

    def forget(self, attrname, *ids):
        '''Issues several "forget related object" commands for the attribute
        named `attrname`.

        '''
        assert self._is_many2many(attrname)
        from .orm import FORGET_RELATED
        cmds = self._commands.setdefault(attrname, [])
        cmds.extend(FORGET_RELATED(id) for id in ids)

    def forgetall(self, attrname):
        '''Issues a "forget all related objects" command for the attribute
        named `attrname`.

        '''
        assert self._is_many2many(attrname)
        from .orm import UNLINKALL_RELATED
        cmds = self._commands.setdefault(attrname, [])
        cmds.append(UNLINKALL_RELATED())

    def replace(self, attrname, *ids):
        '''Issues a "replace related objects" command for the attribute named
        `attrname`.

        '''
        assert self._is_many2many(attrname)
        from .orm import REPLACEWITH_RELATED
        cmds = self._commands.setdefault(attrname, [])
        cmds.append(REPLACEWITH_RELATED(*ids))

    def add(self, attrname, *ids):
        '''Issues several "link to" commands to the attribute named
        `attrname`.

        '''
        from .orm import LINK_RELATED
        assert self._is_many2many(attrname)
        cmds = self._commands.setdefault(attrname, [])
        cmds.extend(LINK_RELATED(id) for id in ids)


class ORMWriter(_BaseWriter):
    '''A context manager that eases writing objects with the Odoo's ORM.

    Two possible signatures:

    - ``get_writer(ModelRecordSet)``
    - ``get_writer(obj, cr, uid, ids, *, context=None)``

    .. warning:: `context` is a keyword-only argument.

    In the first case, ``ModelRecordSet`` should be a record set obtained from
    new API ``browse`` method. If the record set is empty, raise a ValueError.

    The second case is for the old API.

    Usage::

       with get_writer(obj, cr, uid, ids, context=context) as writer:
          writer.update(attr1=val1, attr2=val2, attr3=None)
          writer.update(attr4=val4)
          writer.add(many2manycolumn, id1, id2, id3)
          writer.forget(many2manycolumn, id4, id5)

    At the end of the `with` sentence the equivalent ``obj.write()`` method
    will be called.

    .. warning:: Non-magical disclaimer.

       The sole purpose of writers is to ease the writing of write sentences.
       If the OpenERP does not understand the commands you produce is not our
       fault.

    '''
    def __init__(self, *args, **kwargs):
        try:
            if kwargs or len(args) > 1:
                context = kwargs.pop('context', None)
                _self, cr, uid, ids = args
                recordset = _self.browse(cr, uid, ids, context=context)
            elif len(args) == 1:
                recordset = args[0]
                if not recordset.ids:
                    raise ValueError('Invalid record set for get_writer')
            else:
                raise TypeError('Invalid signature for get_writer')
        except (KeyError, ValueError):
            raise TypeError('Invalid signature for get_writer')
        super(ORMWriter, self).__init__(recordset)

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type or exc_value:
            return False
        # Both the new API and old API return True/False, no downgrade needed.
        self.result = self.model.write(self.commands)


class ORMCreator(_BaseWriter):
    '''A context manager that eases creation of objects with Odoo's ORM.

    Two possible signatures:

    - ``get_creator(ModelRecordSet)``
    - ``get_creator(obj, cr, uid, ids, *, context=None)``

    .. warning:: `context` is a keyword-only argument.

    In the first case, ``ModelRecordSet`` should be a record set obtained from
    new API ``browse`` method.  The recordset may be empty.

    The second case is for the old API (we do the browse).

    This is similar to `get_writer`:func: but issues a 'create' call at the
    exit of the ``with`` block.  The context manager yields a creator that,
    after creating the object, stores the result in the attribute ``result``.
    If you use the new API call style, the ``result`` will be the record
    created.  If you use the old API call style, you'll get the id of created
    record.

    '''
    def __init__(self, *args, **kwargs):
        try:
            if kwargs or len(args) > 1:
                context = kwargs.pop('context', None)
                _self, cr, uid = args
                self.downgrade = True
                model = _self.browse(cr, uid, [], context=context)
            elif len(args) == 1:
                self.downgrade = False
                model = args[0]
            else:
                raise TypeError('Invalid signature for get_creator')
        except (KeyError, ValueError):
            raise TypeError('Invalid signature for get_creator')
        super(ORMCreator, self).__init__(model)

    def __exit__(self, exc_type, exc_value, exc_tb):
        if exc_type or exc_value:
            return False
        result = self.model.create(self.commands)
        if self.downgrade:
            # old API used, old API expectations returned
            self.result = result.id
        else:
            self.result = result

    set = _BaseWriter.update



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from itertools import product
import unittest
from hypothesis import strategies as s, given

from xoeuf.osv import expression as expr
from xoeuf.odoo.osv import expression as odoo_expr
from xoeuf.osv.expression import Domain, DomainTree

names = s.text(alphabet='abdefgh', min_size=1, average_size=3)
operators = s.sampled_from(['=', '!=', '<', '>', '<>'])

# Logical connectors with the amount of terms it connects.  Notice that ''
# takes two arguments because it's the same as '&'.
connectors = s.sampled_from([('', 2), ('&', 2), ('!', 1), ('|', 2)])


@s.composite
def terms(draw, values=s.integers(min_value=-10, max_value=10)):
    name = str(draw(names))
    operator = draw(operators)
    value = draw(values)
    return (name, operator, value)


@s.composite
def domains(draw, min_size=1, average_size=4):
    leaves = draw(s.lists(
        terms(),
        min_size=min_size,
        average_size=average_size
    ))
    result = []
    while leaves:
        connector, many = draw(connectors)
        if many <= len(leaves):
            if connector:
                result.append(connector)
            result.extend(leaves[:many])
            leaves = leaves[many:]
        elif len(leaves) == 1:
            result.append(leaves.pop(0))
    return Domain(result)


class TestDomain(unittest.TestCase):
    @given(domains())
    def test_first_normal_form_idempotency(self, domain):
        self.assertEqual(
            domain.first_normal_form,
            domain.first_normal_form.first_normal_form
        )

    @given(domains())
    def test_second_normal_form_idempotency(self, domain):
        self.assertEqual(
            domain.second_normal_form,
            domain.second_normal_form.second_normal_form
        )

    @given(domains())
    def test_versions_equivalency(self, domain):
        # Any domain must be equivalent in all its versions.
        all_versions = [
            domain,
            domain.simplified,
            domain.first_normal_form,
            domain.second_normal_form
        ]
        for version1, version2 in product(all_versions, all_versions):
            self.assertEqual(
                version1,
                version2,
                msg="%r != %r" % (version1, version2)
            )

    @given(domains())
    def test_simplified(self, domain):
        # A simplified domain never start with '&' operator.
        self.assertNotEqual(
            domain.simplified[0],
            expr.AND_OPERATOR,
            msg="%r start with &" % domain.simplified
        )
        # The simplified domain must be the shortest.
        all_versions = [
            domain,
            domain.first_normal_form,
            domain.second_normal_form
        ]
        for version in all_versions:
            self.assertLessEqual(
                len(domain.simplified),
                len(version),
                msg="len(%r) > len(%r)" % (domain.simplified, version)
            )

    @given(domains())
    def test_simplified_idempotency(self, domain):
        self.assertEqual(
            domain.simplified,
            domain.simplified.simplified
        )

    @given(s.lists(domains(), max_size=10, average_size=5))
    def test_expr_replacement(self, domains):
        expr.AND(domains)
        odoo_expr.AND(domains)

    def test_eq(self):
        A = Domain([
            ('field_x', '=', 1)
        ])
        # A == A
        self.assertTrue(A == A)

        # ~A == ~A
        self.assertTrue(~A == ~A)

        # (x = 1)  == ! (x <> 1)
        A1 = Domain([
            '!',
            ('field_x', '<>', 1)
        ])
        self.assertTrue(A == A1)

    def test_implies(self):
        # Basic implications.
        # -------------------

        A = Domain([
            ('field_x', '=', 1)
        ])
        B = Domain([
            ('field_y', '!=', False)
        ])

        # A => A
        self.assertTrue(A.implies(A))

        # A => A | B
        self.assertTrue(A.implies(A | B))

        # A & B => A
        self.assertTrue((A & B).implies(A))

        # A & B => A | B
        self.assertTrue((A & B).implies(A | B))

        self.assertEqual(Domain.AND(list(A), B), (A & B))
        self.assertEqual(expr.AND([list(A), B]), (A & B))

        # Inter terms implications.
        # -------------------------

        # (x = 1)  => (x = 1)
        A = Domain([
            ('field_x', '=', 1)
        ])
        A1 = Domain([
            ('field_x', '=', 1)
        ])
        self.assertTrue(A.implies(A1))

        # (x != 1)  => (x <> 1)
        A = Domain([
            ('field_x', '!=', 1)
        ])
        A1 = Domain([
            ('field_x', '<>', 1)
        ])
        self.assertTrue(A.implies(A1))

        # (x > 1) => (x > 0)
        A = Domain([
            ('field_x', '>', 10)
        ])
        A1 = Domain([
            ('field_x', '>', 1)
        ])
        self.assertTrue(A.implies(A1))

        # (x in (1,2,3)) => (x in (1,2,3,4))
        A = Domain([
            ('field_x', 'in', (1, 2, 3))
        ])
        A1 = Domain([
            ('field_x', 'in', (1, 2, 3, 4))
        ])
        self.assertTrue(A.implies(A1))
        #

        # Compound implications.
        # ----------------------
        # A => B & C iff (A => B) & (A => C)
        A = Domain([
            ('field_x', 'in', (1, 4))
        ])
        B = Domain([
            ('field_x', 'in', (1, 2, 4))
        ])
        C = Domain([
            ('field_x', 'in', (1, 3, 4))
        ])
        self.assertTrue(A.implies(B))
        self.assertTrue(A.implies(C))
        self.assertTrue(A.implies(B & C))

        # More complex implications
        # -------------------------

        x = Domain([
            '|',
            ('field_y', '!=', False),
            ('field_x', '=', 'value'),
            ('field_z', 'in', (1,))
        ])
        y = Domain([
            '|',
            ('field_y', '!=', False),
            ('field_x', '=', 'value'),
            ('field_z', 'in', (1, 2, 3)),
            '|',
            ('field_x', '=', 'value'),
            ('field_y', '!=', False)
        ])
        assert x.implies(y)

        x = Domain([
            ('field_y', '!=', False),
            ('field_z', 'in', (1,)),
            ('field_z', 'in', (1, 2)),
            '|',
            ('field_y', '!=', False),
            '|',
            ('field_z', 'in', (1,)),
            ('field_z', 'in', (2,))
        ])
        y = Domain([
            ('field_y', '!=', False),
            ('field_z', 'in', (1, 2, 3))
        ])
        assert x.implies(y)

    def test_walk(self):
        y = Domain([
            '|',
            ('a', '!=', False),
            ('b', '=', 'value'),
            ('c', 'in', (1, 2, 3)),
            '|', '|', '!',
            ('d', '=', 'value'),
            ('f', '!=', False),
            ('g', '=', 'h')
        ])
        DomainTree(y.second_normal_form)
        # expected = [
        #     ('TERM', ('c', 'in', (1, 2, 3))),
        #     ('TERM', ('g', '=', 'h')),
        #     ('TERM', ('d', '!=', 'value')),
        #     ('TERM', ('f', '!=', False)),
        #     ('OPERATOR', '|'),
        #     ('OPERATOR', '|'),
        #     ('TERM', ('a', '!=', False)),
        #     ('TERM', ('b', '=', 'value')),
        #     ('OPERATOR', '|'),
        #     ('OPERATOR', '&'),
        #     ('OPERATOR', '&')
        # ]
        # self.assertEqual(expected, list(tree.walk()))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

# This is the implementation of the signals.  The 'signals' module remains the
# API but we're porting this to 'openerp.signals'.

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import logging
logger = logging.getLogger(__name__)
del logging

from xoeuf.odoo import api, models
from xoutil.objects import temp_attributes


class HookDefinition(object):
    def __init__(self, action=None, doc=None):
        self.hooks = []
        self.action = action
        self.__doc__ = doc

    def __repr__(self):
        return '<Signal(%r)>' % self.action

    def connect(self, hook, sender=None, require_registry=True,
                framework=False):
        """Connect hook.

        :param hook: A function or an instance method which is to receive
                handle the hook.  Hooks must be hashable objects and must be
                able to accept keyword arguments.

        :param sender: The sender(s) to which the receiver should
                respond. Must either a model name, list of model names or None
                to receive events from any sender.

        :param require_registry: If True the receiver will only be called if a
               the actual `sender` of the signal has a ready DB registry.

        :keyword framework: Set to True to make this a `framework-level hook
                            <FrameworkHook>`:class:.

        :return: receiver

        """
        if not isinstance(sender, (list, tuple)):
            sender = [sender]
        HookClass = Hook if not framework else FrameworkHook
        for s in sender:
            lookup_key = (_make_id(hook), _make_model_id(s))
            if not any(lookup_key == r_key for r_key, _ in self.hooks):
                self.hooks.append(
                    (lookup_key,
                     HookClass(hook, sender=s,
                               require_registry=require_registry))
                )
        return hook

    def disconnect(self, hook=None, sender=None):
        """Disconnect hook from sender.

        :param hook: The registered hook to disconnect.

        :param sender: The registered sender to disconnect

        """
        key = (_make_id(hook), _make_model_id(sender)), hook
        if key in self.hooks:
            self.hooks.remove(key)

    def has_listeners(self, sender=None):
        return bool(self.live_hooks(sender))

    def live_hooks(self, sender):
        """Filter sequence of hooks to get resolved (live hooks).

        Live hook are defined as those that are installed in the DB and apply
        to the given sender.  Framework-level hooks are always live.

        """
        if isinstance(sender, models.Model):
            registry_ready = sender.pool.ready
        else:
            registry_ready = False
        result = []
        for _, hook in self.hooks:
            if hook.is_installed(sender) and hook.matches(sender):
                if registry_ready or not hook.require_registry:
                    logger.debug(
                        'Accepting hook %s as live',
                        hook,
                        extra=dict(
                            registry_ready=registry_ready,
                            registry_required=hook.require_registry,
                        )
                    )
                    result.append(hook)
        return result


class Signal(HookDefinition):
    """Base class for all signals

    """
    def send(self, sender, **kwargs):
        """Send signal from sender to all connected receivers.

        If any receiver raises an error, the error propagates back through
        send, terminating the dispatch loop, so it is quite possible to not
        have all receivers called if a raises an error.

        If the `sender` is not an Odoo model any `receiver`:func: that
        requires a ready registry will no be called.

        :param sender: The sender of the signal either a model or None.

        :param kwargs: Named arguments which will be passed to receivers.

        :return: Returns a list of tuple pairs [(receiver, response), ... ].

        """
        responses = []
        if not self.hooks:
            return responses
        for hook in self.live_hooks(sender):
            response = hook(sender, self, **kwargs)
            responses.append((hook, response))
        return responses

    def safe_send(self, sender, catched=(Exception, ), thrown=None, **kwargs):
        """Send signal from sender to all connected receivers catching errors.

        :param sender: The sender of the signal either a model or None.

        :keyword catched: A (tuple of ) exception to safely catch. The default
                          is ``(Exception, )``.

        :keyword thrown: A (tuple of) exceptions to re-raise even though in
                 `catched`.  The default is not to re-raise any error.

        :return: Returns a list of tuple pairs [(receiver, response), ... ].

        All remaining keyword arguments are passed to receivers.

        If any receiver raises an error (specifically any subclass of
        Exception), the error instance is returned as the result for that
        receiver.

        """
        from celery.exceptions import SoftTimeLimitExceeded
        responses = []
        if thrown and not isinstance(thrown, (list, tuple)):
            thrown = (thrown, )
        if not self.hooks:
            return responses
        for hook in self.live_hooks(sender):
            try:
                response = hook(sender, self, **kwargs)
            except SoftTimeLimitExceeded:
                raise
            except catched as err:
                if thrown and isinstance(err, thrown):
                    # Don't log: I expect you'll log where you actually catch
                    # it.
                    raise
                else:
                    logger.exception(err)
                    responses.append((hook, err))
            else:
                responses.append((hook, response))
        return responses


class Wrapping(HookDefinition):
    def perform(self, method, sender, *args, **kwargs):
        livewrappers = self.live_hooks(sender)
        wrappers = []
        for wrapper in livewrappers:
            try:
                w = wrapper(sender, self, *args, **kwargs)
                try:
                    next(w)
                except StopIteration:
                    logger.error(
                        'Wrapper %s failed to yield once',
                        wrapper
                    )
                else:
                    wrappers.append(w)
            except Exception:
                logger.exception('Unexpected error in wrapper')
        result = method(sender, *args, **kwargs)
        for wrapper in wrappers:
            try:
                wrapper.send(dict(result=result))
                logger.error('Wrapper %s failed to yield only once')
            except StopIteration:
                pass
            except Exception:
                logger.exception('Unexpected error in wrapper')
        return result


class Hook(object):
    '''Wraps a hook function, so that we can store some metadata.'''
    def __init__(self, func, **kwargs):
        from xoutil.objects import smart_copy
        hash(func)  # Fail if func is not hashable
        self.func = func
        smart_copy(
            kwargs,
            self.__dict__,
            defaults={'require_registry': True, 'sender': None}
        )

    def __repr__(self):
        return '<Hook for %r>' % self.func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    # So that the original receiver compares equal to this wrapper.  This
    # avoids having the same function being registered twice.
    def __hash__(self):
        return hash(self.func)

    def __eq__(self, other):
        return self.func == other

    def matches(self, sender):
        key = _make_model_id(sender)
        return not self.sender or key == _make_model_id(self.sender)

    def is_installed(self, sender):
        '''Check whether this receiver is installed in the DB of `sender`.

        If the sender does not have an 'env' attribute it will considered
        installed as well (this is for dispatching outside the Odoo model
        framework.)

        '''
        from xoeuf.modules import get_object_module
        module = get_object_module(self.func, typed=True)
        env = getattr(sender, 'env', None)
        if module and env:
            mm = env['ir.module.module'].sudo()
            query = [('state', '=', 'installed'), ('name', '=', module)]
            with _no_signalling(pre_search), _no_signalling(post_search):
                return bool(mm.search(query))
        else:
            return True


class FrameworkHook(Hook):
    '''A hook that is defined in a framework-level module.

    Framework-level hooks are always considered installed in any DB.  So you
    should be careful no requiring addon-level stuff.

    '''
    def is_installed(self, sender):
        '''Check whether this receiver is installed in the DB of `sender`.'''
        return True


def receiver(signal, **kwargs):
    """A decorator for connecting receivers to signals.

    :param signal: Either a single signal or a list of signals.

    :keyword require_registry: If set to True the receiver will only be called
             if the Odoo DB registry is ready.

    :keyword framework: Set to True to make this a `framework-level receiver
                        <FrameworkHook>`:class:.

    Used by passing in the signal (or list of signals) and keyword arguments
    to connect::

        @receiver(post_save, sender='my.model')
        def signal_receiver(sender, **kwargs):
            ...

        @receiver([post_save, post_delete], sender='my.model')
        def signals_receiver(sender, **kwargs):
            ...

    """
    def _decorator(func):
        if isinstance(signal, (list, tuple)):
            for s in signal:
                s.connect(func, **kwargs)
        else:
            signal.connect(func, **kwargs)
        return func
    return _decorator


def wrapper(wrapping, **kwargs):
    """A decorator for connecting wrappers.

    :param wrapping: Either a single wrapping or a list of them.

    :keyword require_registry: If set to True the wrapper will only be called
             if the Odoo DB registry is ready.

    :keyword framework: Set to True to make this a `framework-level wrapper
                        <FrameworkHookData>`:class:.

    Example::

        @wrapper(write_wrapper, sender='my.model')
        def _do_what_i_mean(sender, **kwargs):
            yield

    Wrappers must yield exactly once.  The code before the ``yield`` runs
    before the original method.  Any errors in the code of the wrapper are
    logged and ignored: the original method always runs.

    Standard wrappers wrap the pre/post signals.  So wrappers may be affected
    by the effects in post signals.

    """
    return receiver(wrapping, **kwargs)


def _no_signalling(signal):
    '''Context manager that temporarily stop the signal from being called.

    Basically, we disconnect all receivers from `signal` within the scope of
    the context manager.

    '''
    return temp_attributes(signal, {'hooks': []})


def mock_replace(hook, func, **replacement_attrs):
    '''Mock a hook.

    Example::

       with mock_replace(post_create, receiver) as mock:
          # do things that should trigger post_create
          assert mock.called

    If the `receiver` is not connected to the  to `hook` it will still return
    a mock, that should never be called.

    '''
    import contextlib
    from xoutil.symbols import Undefined
    try:
        from unittest.mock import MagicMock
    except ImportError:
        from mock import MagicMock

    @contextlib.contextmanager
    def _hidden_patcher():
        mock = MagicMock(func, **replacement_attrs)
        registry = hook.hooks
        pos = find(func, registry, extract=snd)
        if pos is not Undefined:
            _, receiver_holder = registry[pos]
            previous_receiver = receiver_holder.func
            receiver_holder.func = mock
            previous_is_installed = receiver_holder.is_installed
            receiver_holder.is_installed = is_installed
        else:
            receiver_holder = None
        try:
            yield mock
        finally:
            if receiver_holder is not None:
                receiver_holder.func = previous_receiver
                receiver_holder.is_installed = previous_is_installed

    def snd(pair):
        _, res = pair
        return res

    def find(what, l, extract=None):
        if not extract:
            extract = lambda x: x  # noqa: E731
        ll = [extract(x) for x in l]
        try:
            return ll.index(what)
        except ValueError:
            return Undefined

    def is_installed(sender):
        '''Mocked is installed'''
        from xoeuf.modules import get_object_module
        module = get_object_module(func, typed=True)
        env = getattr(sender, 'env', None)
        if module and env:
            mm = env['ir.module.module'].sudo()
            query = [('state', '=', 'installed'), ('name', '=', module)]
            with _no_signalling(pre_search), _no_signalling(post_search):
                return bool(mm.search(query))
        else:
            return True

    return _hidden_patcher()


def _make_id(target):
    if hasattr(target, '__func__'):
        return (id(target.__self__), id(target.__func__))
    return id(target)


def _issubcls(which, Class):
    return isinstance(which, type) and issubclass(which, Class)


def _make_model_id(sender):
    '''Creates a unique key for 'senders'.

    Since Odoo models can be spread across several classes we can't simply
    compare by class object.  So if 'sender' is a BaseModel (subclass or
    instance), the key will be the same for all classes targeting the same
    model.

    '''
    BaseModel = models.BaseModel
    if isinstance(sender, BaseModel) or _issubcls(sender, BaseModel):
        return sender._name
    else:
        return sender


# **************SIGNALS DECLARATION****************
pre_fields_view_get = Signal('pre_fields_view_get')
post_fields_view_get = Signal('post_fields_view_get')

pre_search = Signal('pre_search', '''
Signal sent when the 'search' method is to be invoked.

Arguments:

:param sender: The recordset where the 'search' was called.

:keyword query: The domain (argument `args` in Odoo's search method).
                This will be `list`:class: that can be modified in place, to
                customize the search.  But notice that the order in which the
                receivers are called is not defined.

:keyword kw_args: The rest of the arguments to 'search'.  We make it a dict as
                  if called by keyword.

''')


post_search = Signal('post_search', '''
Signal sent after the 'search' method was invoked.

Arguments:

:param sender: The recordset where the 'search' was called.

:keyword query: The domain (argument `args` in Odoo's search method).
                This will be `list`:class: that can be modified in place, to
                customize the search.

:keyword kw_args: The rest of the arguments to 'search'.  We make it a dict as
                  if called by keyword.

:keyword result: The result of the actual 'search'.

''')


pre_create = Signal('pre_create', '''
Signal sent when the 'create' method is to be invoked.

If a receiver raises an error the create is aborted, and post_create won't be
issued.  The error is propagated.

Arguments:

:param sender: The recordset where the 'create' was called.

:keyword values: The values passed to 'create'.

''')

post_create = Signal('post_create', '''
Signal sent when the 'create' method has finished but before data is committed
to the DB.

If the 'create' raises an error no receiver is invoked.

If a receiver raises an error, is trapped and other receivers are allowed
to run.  However if the error renders the cursor unusable, other receivers
and the commit to DB may fail.

If a receiver raises an error the create is halted and the error is
propagated.

Arguments:

:param sender: The recordset where the 'create' was called.

:keyword result: The result of the call to 'create'.
:keyword values: The values passed to 'create'.
''')

pre_write = Signal('pre_write', '''
Signal sent when the 'write' method of model is to be invoked.

If a receiver raises an error the write is aborted and 'post_write' is not
sent.  The error is propagated.

Arguments:

:param sender: The recordset sending the signal.

:keyword values: The values passed to the write method.

''')
post_write = Signal('post_write', '''
Signal sent after the 'write' method of model was executed.

If 'write' raises an error no receiver is invoked.  If a receiver raises an
error is trapped (see `safe_send`) and other receivers are allowed to run.
However, if the error renders the cursor unusable other receivers may fail
and the write may fail to commit.

Arguments:

:param sender: The recordset sending the signal.

:keyword result: The result from the write method.

:keyword values: The values passed to the write method.

''')

pre_unlink = Signal('pre_unlink', '''
Signal sent when the 'unlink' method of model is to be invoked.

If a receiver raises an error unlink is aborted and 'post_unlink' is not
called.  The error is propagated.

Arguments:

:param sender: The recordset sending the signal.

''')

post_unlink = Signal('post_unlink', '''
Signal sent when the 'unlink' method of a model was executed.

If the 'unlink' raises an error no receiver is invoked.  If a receiver
raises an error is trapped (see `safe_send`) other receivers are allowed to
run.  However, if the error renders the cursor unusable other receivers may
fail and the unlink may fail to commit.

Arguments:

:param sender: The recordset sending the signal.

:keyword result:  The result from the unlink method.

''')

pre_save = [pre_create, pre_write, pre_unlink]
post_save = [post_create, post_write, post_unlink]


# **************SIGNALS SEND****************
super_fields_view_get = models.BaseModel.fields_view_get
super_create = models.BaseModel.create
super_write = models.BaseModel.write
super_unlink = models.BaseModel.unlink
super_search = models.BaseModel.search


@api.model
def _fvg_for_signals(self, view_id=None, view_type='form',
                     toolbar=False, submenu=False):
    kwargs = dict(
        view_id=view_id,
        view_type=view_type,
        toolbar=toolbar,
        submenu=submenu
    )
    pre_fields_view_get.send(sender=self, **kwargs)
    result = super_fields_view_get(self, **kwargs)
    post_fields_view_get.safe_send(sender=self, result=result, **kwargs)
    return result


@api.model
@api.returns('self', lambda value: value.id if value else value)
def _create_for_signals(self, vals):
    pre_create.send(sender=self, values=vals)
    res = super_create(self, vals)
    post_create.safe_send(sender=self, result=res, values=vals)
    return res


@api.multi
def _write_for_signals(self, vals):
    pre_write.send(self, values=vals)
    res = super_write(self, vals)
    post_write.safe_send(self, result=res, values=vals)
    return res


@api.multi
def _unlink_for_signals(self):
    pre_unlink.send(self)
    res = super_unlink(self)
    post_unlink.safe_send(self, result=res)
    return res


write_wrapper = Wrapping('write_wrapper', '''\
Wraps the `write` method.

''')


@api.multi
def _write_for_wrappers(self, vals):
    return write_wrapper.perform(_write_for_signals, self, vals)


@api.model
@api.returns(*super_search._returns)
def _search_for_signals(self, args, offset=0, limit=None, order=None, count=False):
    query = list(args)
    kw_args = dict(offset=offset, limit=limit, order=order, count=count)
    pre_search.send(self, query=query, kw_args=kw_args)
    result = super_search(self, query, **kw_args)
    post_search.safe_send(self, query=query, kw_args=kw_args, result=result)
    return result


models.BaseModel.fields_view_get = _fvg_for_signals
models.BaseModel.create = _create_for_signals
models.BaseModel.unlink = _unlink_for_signals
models.BaseModel.write = _write_for_wrappers
models.BaseModel.search = _search_for_signals



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Initialize the OpenERP logger, and provides a function to smartly get a
logger.

'''
from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoeuf.odoo.netsvc import init_logger


DEFAULT_LOGGER_NAME = str('xoeuf')


def get_logger(name=None):
    '''If a `name` is given, is normally getter, otherwise look for the upper
    module name.

    '''
    from logging import getLogger
    if not name:
        import sys
        frame = sys._getframe(1)
        name = frame.f_globals.get('__name__', DEFAULT_LOGGER_NAME)
    return getLogger(str(name))


init_logger()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import sys
import os
from setuptools import setup, find_packages

# Import the version from the release module
project_name = 'xoeuf'
_current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.join(_current_dir, project_name))
from release import VERSION as version


def safe_read(*paths):
    fname = os.path.join(_current_dir, *paths)
    try:
        with open(fname, 'r') as fh:
            return fh.read()
    except (IOError, OSError):
        return ''


setup(name=project_name,
      version=version,
      description="Basic utilities for OpenERP Open Object Services",
      long_description=safe_read('docs', 'readme.txt'),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
          "Programming Language :: Python",
          "Intended Audience :: Developers",
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',  # noqa
          'Development Status :: 4 - Beta',
      ],
      keywords='openerp open-object server library'.split(),
      author='Merchise Autrement [~º/~]',
      author_email='',
      url='http://www.merchise.org/',
      license='GPL',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'decorator>=4.0,<4.2',
          'xoutil>=1.9.0,<2.0; python_version < "3.4"',
          'xoutil>=1.9.0; python_version >= "3.4"',
          'ipython<6',
          'raven>=5.8.0',
          'raven-sanitize-openerp',
          'enum34; python_version < "3.4"',
          'celery>=4.1.0,<5',
      ],
      extra_requires={
          'odoo': [
              'odoo==10.0; python_version < "3.0"',
              'odoo==11.0; python_version >= "3.5"',
          ],
          'test': ['hypothesis>=3.7.0,<4', ],
      },
      entry_points="""
      [console_scripts]
      xoeuf = xoeuf.cli.server:server
      xoeuf_mailgate = xoeuf.cli.mailgate:main

      [xoeuf.addons]
      test_localized_dt = xoeuf.tests.test_localized_dt
      """)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoeuf import models, fields


class A(models.Model):
    _name = 'test.one2one.a'
    name = fields.Char()


class B(models.Model):
    _name = 'test.one2one.b'
    _inherits = {'test.one2one.a': 'a_id'}


class C(models.Model):
    _name = 'test.one2one.c'
    _inherits = {'test.one2one.a': 'a_id'}

    a_id = fields.One2one('test.one2one.a')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoeuf.odoo.tests.common import TransactionCase


class TestFromActiveIds(TransactionCase):
    # I don't really need to create records, since we're not reading anything
    # from the the DB

    def test_self_takes_from_active_ids(self):
        Model = self.env['xoeuf.tests.test_api.model']
        res = Model.with_context(
            active_model=Model._name,
            active_ids=(1, 2, 3, 4)
        ).return_self_ids()
        self.assertEqual(res, [1, 2, 3, 4])

    def test_not_leaked_context(self):
        Model = self.env['xoeuf.tests.test_api.model']
        this = Model.with_context(
            active_model=Model._name,
            active_ids=(1, 2, 3, 4)
        )
        res, res2 = this.return_ids_and_call_method((5, 6), 'return_self_ids')
        self.assertEqual(res, [1, 2, 3, 4])
        self.assertEqual(res2, [5, 6])

    def test_leaked_context(self):
        Model = self.env['xoeuf.tests.test_api.model']
        this = Model.with_context(
            active_model=Model._name,
            active_ids=(1, 2, 3, 4)
        )
        res, res2 = this.leaked_return_ids_and_call_method((5, 6), 'return_self_ids')
        self.assertEqual(res, [1, 2, 3, 4])
        self.assertEqual(res2, res)
